//@ts-check
"use strict";

/**
 * Tree traversal is totally math, right?
 * @param {object} obj Object to traverse
 * @param {string} path Dot-separated path in object
 */
function traverse(obj, path) {
  let result = obj;
  for (const part of path.split(".")) {
    result = result[part];
    if (result == null) return null;
  }
  return result;
}


/**
 * @typedef EaseIn
 * @type {(t: number) => number}
 */

/**
 * @typedef PointLike
 * @type {{x: number, y: number}}
 */

/** @type EaseIn */
const easeInSin = (t) => 1 + Math.sin(Math.PI / 2 * t - Math.PI / 2);
/** @type EaseIn */
const easeOutSin = (t) => Math.sin(Math.PI / 2 * t);
/** @type EaseIn */
const easeInOutSin = (t) => (1 + Math.sin(Math.PI * t - Math.PI / 2)) / 2;
/** @type EaseIn */
const easeInQuad = (t) => t * t;
/** @type EaseIn */
const easeOutQuad = (t) => t * (2 - t);
/** @type EaseIn */
const easeInOutQuad = (t) => t < 0.5 ? 2 * t * t : -1 + (4 - 2 * t) * t;

/** @type EaseIn */
const easeInCubic = (t) => t * t * t;
/** @type EaseIn */
const easeOutCubic = (t) => (--t) * t * t + 1;
/** @type EaseIn */
const easeInOutCubic = (t) => t < 0.5 ? 4 * t * t * t : (t - 1) * (2 * t - 2) * (2 * t - 2) + 1;

const blink = (x) => x % 2 > 1 ? 1 : 0;
/** @param {number} v @param {number} min @param {number} max @returns {number} */
const clamp = (v, min, max) => Math.max(min, Math.min(max, v));
/** @param {number} time @param {number} startDelay @param {number} endDelay @returns {number} */
const progress = (time, startDelay, endDelay) => clamp((time - startDelay) / (endDelay - startDelay), 0, 1);
/** @param {number} a @param {number} b @param {number} t @returns {number} */
const lerp = (a, b, t) => a + (b - a) * t;
/** @param {number} a @param {number} b @param {number} t @returns {number} */
const lerpAngle = (a, b, t) => a + wrapAngle(wrapAngle(b) - wrapAngle(a)) * t;
/** @param {PointLike} a @param {PointLike} b @param {number} t @returns {PIXI.Point} */
const lerpxy = (a, b, t) => new PIXI.Point(a.x + (b.x - a.x) * t, a.y + (b.y - a.y) * t);
/** @param {PointLike} a @param {PointLike} b @returns {number} */
const distxy = (a, b) => Math.sqrt((a.x - b.x) * (a.x - b.x) + (a.y - b.y) * (a.y - b.y));
/** @param {PointLike} a @returns {number} */
const lengthxy = (a) => Math.sqrt(a.x * a.x + a.y * a.y);
/** @param {PointLike} a @param {PointLike} b @returns {number} */
const anglexy = (a, b) => Math.atan2(a.y - b.y, a.x - b.x);

/** @param {number} a @returns {number} */
function wrapAngle(a) {
  while (a < -Math.PI) {
    a += Math.PI * 2;
  }
  while (a > Math.PI) {
    a -= Math.PI * 2;
  }
  return a;
}

let angleComparisonFactor = 20;

/** @param {number} a @param {number} b @returns {number} */
function angleDist(a, b) {
  a = wrapAngle(a);
  b = wrapAngle(b);
  return Math.abs(b - a);
}

/**
 *
 * @param {number} min Lower bound
 * @param {number} max Upper bound
 * @returns {number}
 */
const randFloat = (min, max) => Math.random() * (max - min) + min;

/**
 * 
 * @param {number} min Lower bound (inclusive)
 * @param {number} max Upper bound (inclusive)
 * @returns {number}
 */
const randInt = (min, max) => Math.floor(Math.random() * (max + 1 - min)) + min;

/**
 * Find the intersection between two segments
 * @param {PIXI.Point} ps1
 * @param {PIXI.Point} pe1
 * @param {PIXI.Point} ps1
 * @param {PIXI.Point} pe2
 * @returns {PIXI.Point}
 */
const findSegmentsIntersection = (ps1, pe1, ps2, pe2) => {
  // Get A,B of first line - points : ps1 to pe1
  const A1 = pe1.y - ps1.y;
  const B1 = ps1.x - pe1.x;
  // Get A,B of second line - points : ps2 to pe2
  const A2 = pe2.y - ps2.y;
  const B2 = ps2.x - pe2.x;

  // Get delta and check if the lines are parallel
  const delta = A1 * B2 - A2 * B1;
  if (delta == 0) return null;

  // Get C of first and second lines
  const C2 = A2 * ps2.x + B2 * ps2.y;
  const C1 = A1 * ps1.x + B1 * ps1.y;
  //invert delta to make division cheaper
  const invdelta = 1 / delta;
  // now return the Vector2 intersection point

  const intersection = new PIXI.Point((B2 * C1 - B1 * C2) * invdelta, (A1 * C2 - A2 * C1) * invdelta);

  // check if point is on both the segment
  const eps = 0.1;
  if (ps1.x < pe1.x) {
    if (intersection.x < ps1.x - eps || intersection.x > pe1.x + eps) return null;
  } else {
    if (intersection.x > ps1.x + eps || intersection.x < pe1.x - eps) return null;
  }

  if (ps1.y < pe1.y) {
    if (intersection.y < ps1.y - eps || intersection.y > pe1.y + eps) return null;
  } else {
    if (intersection.y > ps1.y + eps || intersection.y < pe1.y - eps) return null;
  }

  if (ps2.x < pe2.x) {
    if (intersection.x < ps2.x - eps || intersection.x > pe2.x + eps) return null;
  } else {
    if (intersection.x > ps2.x + eps || intersection.x < pe2.x - eps) return null;
  }

  if (ps2.y < pe2.y) {
    if (intersection.y < ps2.y - eps || intersection.y > pe2.y + eps) return null;
  } else {
    if (intersection.y > ps2.y + eps || intersection.y < pe2.y - eps) return null;
  }

  return intersection;
}

/**
 * Find the intersection between a segment and a rectangle
 * @param {PIXI.Point} ps1
 * @param {PIXI.Point} ps2
 * @param {PIXI.Rectangle} rect
 * @returns {PIXI.Point}
 */
const findSegmentRectIntersection = (ps1, ps2, rect) => {
  const topLeft = new PIXI.Point(rect.x, rect.y + rect.height);
  const bottomLeft = new PIXI.Point(rect.x, rect.y);
  const topRight = new PIXI.Point(rect.x + rect.width, rect.y + rect.height);
  const bottomRight = new PIXI.Point(rect.x + rect.width, rect.y);

  let closestIntersection = null;
  let closestIntersectionDistance = Infinity;

  const leftIntersection = findSegmentsIntersection(ps1, ps2, topLeft, bottomLeft);
  if (leftIntersection != null) {
    const leftIntersectionDistance = distxy(ps1, leftIntersection);

    if (leftIntersectionDistance < closestIntersectionDistance) {
      closestIntersection = leftIntersection;
      closestIntersectionDistance = leftIntersectionDistance;
    }
  }

  const rightIntersection = findSegmentsIntersection(ps1, ps2, topRight, bottomRight);
  if (rightIntersection != null) {
    const rightIntersectionDistance = distxy(ps1, rightIntersection);

    if (rightIntersectionDistance < closestIntersectionDistance) {
      closestIntersection = rightIntersection;
      closestIntersectionDistance = rightIntersectionDistance;
    }
  }

  const topIntersection = findSegmentsIntersection(ps1, ps2, topLeft, topRight);
  if (topIntersection != null) {
    const topIntersectionDistance = distxy(ps1, topIntersection);

    if (topIntersectionDistance < closestIntersectionDistance) {
      closestIntersection = topIntersection;
      closestIntersectionDistance = topIntersectionDistance;
    }
  }

  const bottomIntersection = findSegmentsIntersection(ps1, ps2, bottomLeft, bottomRight);
  if (bottomIntersection != null) {
    const bottomIntersectionDistance = distxy(ps1, bottomIntersection);

    if (bottomIntersectionDistance < closestIntersectionDistance) {
      closestIntersection = bottomIntersection;
      closestIntersectionDistance = bottomIntersectionDistance;
    }
  }

  return closestIntersection;
}
